///////////////////////////////////////////////////////////////////////////////
// File:       SymbolTable.java
//
// Assignment: Assignment 2: Hashing and Sorting
// Course:     CS367, Spring 2004
// Date:       3/3/04
//
// Student:    Justin Fiedler
// Lecture:    
// CS Login:   fiedler
///////////////////////////////////////////////////////////////////////////////

import java.util.ArrayList;
import java.util.LinkedList;
import java.io.*;

/**
 * The SymbolTable class is a chain-bucket hash table to store Symbols
 *
 * <p>Bugs: no
 *
 * @author Justin Fiedler
 */
public class SymbolTable
{   private SourceFileTokenizer sft;
	private ArrayList al = new ArrayList();
	private int size;
	private int totalItems = 0;
	
    /**         
     * Constructs an SymbolTable of null elements of size initialSize
     *
     * @param initialSize the size to make the SymbolTable
     */
    public SymbolTable(int initialSize)
    {	size = initialSize;
    	for(int i = 0; i < initialSize; i++)
    	{	al.add(null);
    	}
    }
    
    /**         
     * Gets the size of the SymbolTable
     *
     * @return the size of the SymbolTable
     */
	public int size()
	{	return size;
	}

	/**         
     * Gets a hash code for the sent String s
     *
     * @param s the string to code
     * @return the hash code for String s
     */
	private int hash(String s)
    {	return (s.charAt(0) * s.length() + s.charAt(s.length() - 1)) % size;
    }
    
	/**         
     * Searches for the String s in the SymbolTable and returns a
     * information string if it is found or else a string reporting
     * that it was not found
     *
     * @param s the string to find
     * @return string representing the Symbol if found else "not found"
     */
    public String find(String s)
    {	LinkedList ll = (LinkedList)al.get(hash(s));
    	if(ll != null)
    	{	if(indexOf(s, ll) != -1)
    		{	return ((Symbol)ll.get(indexOf(s, ll))).toString();
    		}
    	}
    	return s + " not found";
    }
    
    
    /**         
     * Searches for the String s in the LinkedList of Symbols "ll" and returns
     * the index of it if found. If a corresponding symbol is not found 
     * returns -1.
     *
     * @param s the string to find
     * @param ll the LinkedList to search
     * @return the index of 's' or -1 if not found
     */
    private int indexOf(String s, LinkedList ll)
    {	for(int i = 0; i < ll.size(); i++)
    	{	Symbol sym = (Symbol)ll.get(i);
    		if(sym.getString().equals(s))
    		{	return i;
    		}
    	}
    	return -1;
	}
    
    /**         
     * Adds a Symbol of name s to the list in its hash location.  If the Symbol
     * already exists in the SymbolTable it increments its count by 1.
     *
     * @param s the name of the Symbol to add
     * @param reserved whether or not the word is from a reserved word file
     */
    private void add(String s, boolean reserved)
    {	int pos = hash(s);
		if(al.get(pos) == null)
    	{	totalItems += reserved?0:1;
    		LinkedList ll = new LinkedList();
    		ll.add(new Symbol(s, reserved, reserved?0:1));
    		al.remove(pos);
    		al.add(pos, ll);
    	}
    	else
    	{	LinkedList ll = (LinkedList)al.get(pos);
    		pos = indexOf(s, ll);
    		if(pos == -1)
    		{	totalItems += reserved?0:1;
    			ll.add(new Symbol(s, reserved, reserved?0:1));
    		}
    		else
    		{	if(((Symbol)ll.get(pos)).getCount() < 1) totalItems++;
    			((Symbol)ll.get(pos)).increment();
    		}
    	}
    }
    
    /**
     * Returns a string of the SymbolTable with elements in their
     * hash order
     *
     * @return A String of showing the contents of each position
     */
    public String toString()
    {	String s = "";
    	for(int i = 0; i < size; i++)
    	{	LinkedList ll = (LinkedList)al.get(i);
    		s += i + ": ";
    		if(ll != null)
    		{	for(int j = 0; j < ll.size(); j++)
	    		{	s += " --> " + ((Symbol)ll.get(j)).toString();
    			}
    		}
    		s += "\n";
    	}
    	return s;
    }
    
    /**
     *  Reads in a file using the SourceFileTokenizer to split into
     *  tokens.  Creates a Symbol for each token and sets the
     *  reserved state.  If the incoming parameter reserved is true,
     *  than this file is a list of reserved words and all tokens
     *  found should be set as reserved words.  If the reserved 
     *  parameter is false, then each token read may or may not be
     *  a reserved word.  It will have be checked against the current
     *  list of reserved words.  Reserved words should be added to 
     *  the SymbolTable, but their count should be 0 until they are
     *  read from a non-reserved word file.
     *
     *  @param filename The name of the file to read.  It is assumed 
     *  to be in the current working directory or be a fully qualified
     *  path name.
     *  @param reserved If this is true, than all words read are 
     *  to be set as reserved words.  If it is false, then each word
     *  read may or may not be a reserved word.
     */
    public void read(String filename, boolean reserved) throws IOException
    {	sft = null;
    	while(sft == null)
	    {	try
	    	{	sft = new SourceFileTokenizer(filename);
	    	}
	    	catch(FileNotFoundException e)
	    	{	System.out.println("File not found. Must enter existing file.");
	    		throw new IOException();
		}
    	}
    	String token;
    	Symbol sym;
    	token = sft.nextSourceToken();
		while (token != null)
		{   add(token, reserved); 
			token = sft.nextSourceToken();
		}
    }

    /**
     *  Returns a list of all Symbol objects in this SymbolTable
     *  in sorted order.
     *
     *  <p>Example:</p>
     *  <pre>  Symbol Table in sorted order: 
     *  class:R 1
     *  i:N 50
     *  public:R 4
     *  toString:N 1
     *  </pre>
     *  <p>where R means it is a reserved word and N means it is 
     *  not a reserved word and the integer that follows is the
     *  number of occurrences of that symbol.  Symbols that did not 
     *  occur in a source file (other than the reserved file)
     *  are not displayed.  The <a href="Symbol.html">Symbol</a>
     *  will produce this format for the output, if the <tt>toString</tt>
     *  method is used for each symbol.</p>
     *
     *  @return A String of characters with one line for each symbol
     *  as described above.
     */
    public String toSortedString()
    {
        // The sorted string being built
        String ss = "Symbol Table in sorted order: \n";
		Symbol[] sym = this.toArray();
		shellSort(sym, 0, sym.length - 1);
		for(int i = 0; i < sym.length; i++)
		{	ss += sym[i].toString() + '\n';
		}
        return ss;
    } // end toSortedString
    
    /**
     * Returns a Symbol array containing the Symbols contained in the 
     * source file
     *
     * @return a Symbol array containing the contents of the SymbolTable
     */
    private Symbol[] toArray()
    {  	Symbol[] sym = new Symbol[totalItems];
    	int count = 0;
    	for(int i = 0; i < size; i++)
    	{	LinkedList ll = (LinkedList)al.get(i);
    		if(ll != null)
    		{	for(int j = 0; j < ll.size(); j++)
	    		{	Symbol s = (Symbol)ll.get(j);
	    			if(s.getCount() > 0)
	    			{	sym[count++] = s;
	    			}
    			}
    		}
    	}
    	return sym;
    }
    
    /**         
     * Sorts an array of Symbols by name between the positions first
     * and last inclusive. (Relies on insertion sort iSort)
     *
     * @param sym the array of symbols to sort
     * @param first the position of the first symbol to be sorted
     * @param last the position of the first symbol to be sorted
     */
    private void shellSort(Symbol[] sym, int first, int last)
    {	int n = last - first + 1;
    	for(int s = n/2; s > 0; s = s/2)
    	{	for(int i = first; i < first + s; i++)
    		{	iSort(sym, i, last, s);
    		}
    	} 
    }
    
    /**         
     * Sorts an array of Symbols by their name between the positions 
     * first and last inclusive.
     *
     * @param sym the array of symbols to sort
     * @param first the position of the first symbol to be sorted
     * @param last the position of the first symbol to be sorted
     * @param s the amount of space between the symbols to compare
     */
    private void iSort(Symbol[] sym, int first, int last, int s)
    {	int j;
    	for(int i = first + s; i <= last; i += s)
    	{	Symbol temp = sym[i];
    		for(j = i - s; (j >= first) && (temp.getString().compareTo(
    										  sym[j].getString()) < 0); j -= s)
    		{	sym[j + s] = sym[j];
    		}
    		sym[j + s] = temp;
    	}
    }
} // end SymbolTable class
